@@ -33,7 +33,8 @@ fhandler_termios::tcinit (bool is_pty_master)
       tc ()->ti.c_iflag = BRKINT | ICRNL | IXON | IUTF8;
       tc ()->ti.c_oflag = OPOST | ONLCR;
       tc ()->ti.c_cflag = B38400 | CS8 | CREAD;
-      tc ()->ti.c_lflag = ISIG | ICANON | ECHO | IEXTEN;
+      tc ()->ti.c_lflag = ISIG | ICANON | ECHO | IEXTEN
+	| ECHOE | ECHOK | ECHOCTL | ECHOKE;
 
       tc ()->ti.c_cc[VDISCARD]	= CFLUSH;
       tc ()->ti.c_cc[VEOL]	= CEOL;
@@ -438,7 +439,7 @@ fhandler_termios::line_edit (const char *rptr, size_t nread, termios& ti,
       if (ti.c_lflag & ECHO)
 	doecho (&c, 1);
       /* Write in chunks of 32 bytes to reduce the number of WriteFile calls
-      	in non-canonical mode. */
+	in non-canonical mode. */
       if ((!iscanon && ralen () >= 32) || input_done)
 	{
 	  int status = accept_input ();
