@@ -14,7 +14,11 @@ details. */
 #define NO_COPY __attribute__((nocommon)) __attribute__((section(".data_cygwin_nocopy")))
 #define NO_COPY_INIT __attribute__((section(".data_cygwin_nocopy")))
 
+#ifdef __cplusplus
 #define EXPORT_ALIAS(sym,symalias) extern "C" __typeof (sym) symalias __attribute__ ((alias(#sym)));
+#else
+#define EXPORT_ALIAS(sym,symalias) __typeof (sym) symalias __attribute__ ((alias(#sym)));
+#endif
 
 #define _WIN32_WINNT 0x0a00
 #define WINVER 0x0a00
@@ -28,6 +32,7 @@ details. */
 
 #include <sys/types.h>
 #include <sys/strace.h>
+#include <sys/smallprint.h>
 
 /* Declarations for functions used in C and C++ code. */
 #ifdef __cplusplus
@@ -223,8 +228,6 @@ void set_ishybrid_and_switch_to_pcon (HANDLE h);
 /* Printf type functions */
 extern "C" void vapi_fatal (const char *, va_list ap) __attribute__ ((noreturn));
 extern "C" void api_fatal (const char *, ...) __attribute__ ((noreturn));
-int __small_sprintf (char *dst, const char *fmt, ...);
-int __small_vsprintf (char *dst, const char *fmt, va_list ap);
 int __small_swprintf (PWCHAR dst, const WCHAR *fmt, ...);
 int __small_vswprintf (PWCHAR dst, const WCHAR *fmt, va_list ap);
 void multiple_cygwin_problem (const char *, uintptr_t, uintptr_t);
